import raypyng
from .rml import RMLFile
from .rml import ObjectElement,ParamElement
import itertools
import os 
import numpy as np

from .runner import RayUIAPI,RayUIRunner
from .recipes import SimulationRecipe
from .multiprocessing import RunPool
from .postprocessing import PostProcess

################################################################
class SimulationParams():
    """The entry point of the simulation parameters.
    
    A class that takes care of the simulations parameters, 
    makes sure that they are written correctly,
    and returns the the list of simulations that is requested by the user.

    Args:
        rml (RMLFile/string, optional): string pointing to an rml 
                                        file with the beamline template, 
                                        or an RMLFile class object. 
                                        Defaults to None.
        param_list (list, optional): list of dictionaries containing the 
                                     parameters and values to simulate. 
                                     Defaults to None.            
    """    
    def __init__(self, rml=None, param_list=None,**kwargs) -> None:
        """ 
        Args:
            rml (RMLFile/string, optional): string pointing to an rml file with the beamline template, or an RMLFile class object. Defaults to None.
            param_list (list, optional): list of dictionaries containing the parameters and values to simulate. Defaults to None.
        """        
        if rml is not None:
            if isinstance(rml,RMLFile):
                self._rml = rml
            else: # assume that parameter is the file name as required for RMLFile
                self._rml = RMLFile(rml,**kwargs)
        else:
            raise Exception("rml file must be defined")
        
        self.param = param_list

        
    @property 
    def rml(self):
        """RMLFile object instantiated in init
        """        
        return self._rml

    @property
    def params(self):   
        """The parameters to scan, as a list of dictionaries.
        For each dictionary the keys are the parameters elements of the beamline, and the values are the 
        values to be assigned.
        """            
        return self.param

    @params.setter
    def params(self,value):
        # self.param must be a list
        if not isinstance(value, list) == True:
            raise AssertionError('params must be a list')
        # every element in the list must be a dictionary
        for d in value:
            if not isinstance(d, dict):
                raise AssertionError('The elements of params must be dictionaries')
        # the items permitted types are:
        # int,float,str,np.array
        # a list of the types above
        # the output of range, that I still did not understand what it is
        # in any case at the end we want to have either
        # a list or a numpy array
        for d in value:
            for k in d.keys():
                if not isinstance(k,ParamElement):
                    raise AssertionError('The keys of the dictionaries must be instance of ParamElement, while ', k, 'is a ', str(type(k)))
                if isinstance(d[k], (list)):
                    pass
                elif isinstance(d[k], (float, int, str)):
                    d[k] = [d[k]]
                else: # this works with range output, but I would like to capture the type..
                    try:
                        d[k] = list(d[k])
                    except TypeError:
                        raise Exception('The only permitted type are: int, float, str, list, range, np.array, check',d[k]) 
        self.param = value
  
    def _check_param(self):
        """Check that self.param is a list of dictionaries, and convert the 
        items of the dictionaries to lists, otherwise raise an exception.
        """        
        
    def _extract_param(self, verbose:bool=False):
        """Parse self.param and extract dependent and independent parameters

        Args:
            verbose (bool, optional): If True print the returned objects. Defaults to False.

        Returns:
            self.ind_param_values (list): indieendent parameter values
            self.ind_par (list): independent parameters
            self.dep_param_dependency (dict): dictionary of dependencies
            self.dep_value_dependency (list): dictionaries of dependent values
            self.dep_par (list): dependent parameters

        """                 
        self.ind_param_values = []
        self.ind_par = []
        self.dep_param_dependency = {}
        self.dep_value_dependency = []
        self.dep_par = []
        # loop over the list of dictionaries
        for par in self.param:
            keys_par = list(par.keys())
            # if there is more than one key, we have an indipendent parameter
            # and one or more dependent parameters
            # the dependent param are keys, value and dependency are stored
            if len(keys_par) > 1:
                index_param  = 0
                for dep_param in keys_par:
                    if dep_param != keys_par[0]:
                        if len(par[dep_param]) != len(par[keys_par[0]]):
                            dep_param_string = dep_param.get_full_path().lstrip("lab.beamline.")
                            indep_param_string = keys_par[0].get_full_path().lstrip("lab.beamline.")
                            raise AssertionError ('The parameter {} has {} values, but since it depends on {} it \
                                should have the same number of values, {}'.format(dep_param_string,
                                                                                len(par[dep_param]),
                                                                                indep_param_string,
                                                                                len(par[keys_par[0]])))
                        index_values = 0
                        self.dep_param_dependency[dep_param] = keys_par[0]
                        for dep_value in par[dep_param]:
                            if index_values == 0:
                                self.dep_value_dependency.append({par[keys_par[0]][index_values]:dep_value})
                            else:
                                self.dep_value_dependency[index_param][par[keys_par[0]][index_values]]=dep_value
                            index_values += 1
                        index_param += 1
            # here we deal with the indipendent parameters
            self.ind_param_values.append(par[keys_par[0]])
            self.ind_par.append(keys_par[0])
            self.dep_par = list(self.dep_param_dependency.keys())
        if verbose:
            print('###########################################')
            print('self.ind_param_values', self.ind_param_values)
            print('self.ind_par', self.ind_par)
            print('###########################################')
            print('self.dep_param_dependency', self.dep_param_dependency)
            print('self.dep_value_dependency',self.dep_value_dependency)
        return (self.ind_param_values,self.ind_par,self.dep_param_dependency,self.dep_value_dependency,self.dep_par)

    def _make_dictionary(self, keys, items):
        d = {}
        for v,k in enumerate(keys):
            # it follows a dirty little trick to ensure that all the dictionry keys
            # are different
            k.cdata=v+3987.3423421534563632
            d.update({k:items[v]})
        return d

    def params_list(self, obj=None):
        result = []
        for i in self.simulations_param_list:
            result.append(self._make_dictionary(self.param_to_simulate, i))
        return result

    def _calc_loop(self, verbose:bool=True):
        """Calculate the simulations loop

        Returns:
            self.param_to_simulate (list): idependent and dependent parameters
            self.simulations_param_list (list): parameters values for each simulation loop
        """                
        self.param_to_simulate = self.ind_par + self.dep_par
        self.simulations_param_list = []
        # here we arrange the indipendent parameters in a grid
        self.loop = list(itertools.product(*self.ind_param_values))
        
        # work out where are the parameters on which the dependent parameters depends
        # make a copy of the dependency dictionary and replace the items with the index in the second step
        self.dep_param_dependency_index = []
        for ind,par in enumerate(self.dep_param_dependency.values()):
            index_par = self.ind_par.index(par)
            self.dep_param_dependency_index.append(index_par)

        # here we add the dependent parameters
        for count, loop in enumerate(self.loop):
            for ind,par in enumerate(self.dep_param_dependency.keys()):
                to_add = (self.dep_value_dependency[ind][loop[self.dep_param_dependency_index[ind]]],)
                loop = loop + to_add
            self.simulations_param_list.append(loop)
        self.par = self.ind_par + self.dep_par
        if verbose:
            print('You have defined:')
            print(len(self.ind_par), ' independent parameters')
            print(len(self.dep_par), ' dependent parameters or set parameters')
            print(len(self.simulations_param_list), ' simulations')
        return (self.param_to_simulate, self.simulations_param_list)
    
    def _check_if_enabled(self, param):
        """Check if a parameter is enabled

        Args:
            param (RML object): an parameter to simulate

        Returns:
            (bool): True if the parameter is enabled, False otherwise
        """        
        return param.enabled=='T'
    
    def _enable_param(self, param):
        """Set enabled to True in a beamline object, and auto to False

        Args:
            param (RML object): beamline object
        """        
        if not self._check_if_enabled(param):
            param.enabled = 'T'
        try:
            param.auto = 'F'
        except AttributeError:
            pass


    def _write_value_to_param(self, param, value):
        """Write a value to a parameter, making sure enable is T 
        and auto is F

        Args:
            param (RML object): beamline object
            value (str,int,float): the value to set the beamline object to
        """        
        self._enable_param(param)
        if not isinstance(value,str):
            value = str(value)
        param.cdata = value

################################################################
class Simulate():
    """A class that takes care of performing the simulations with RAY-UI 

    Args:
        rml (RMLFile/string, optional): string pointing to an rml file with 
                                        the beamline template, or an RMLFile 
                                        class object. Defaults to None.
        hide (bool, optional): force hiding of GUI leftovers, xvfb needs 
                               to be installed. Defaults to False.
        ray_path (str, optional): the path to the RAY-UI installation folder. 
                                  If None, the program will look for RAY-UI in 
                                  the standard installation paths. 
    """
    def __init__(self, rml=None, hide=False,ray_path=None,**kwargs) -> None:
        """Initialize the class with a rml file
        Args:
            rml (RMLFile/string, optional): string pointing to an rml file with 
                                            the beamline template, or an RMLFile 
                                            class object. Defaults to None.
            hide (bool, optional): force hiding of GUI leftovers, xvfb needs 
                                   to be installed. Defaults to False.
            ray_path (str, optional): the path to the RAY-UI installation folder. 
                                      If None, the program will look for RAY-UI in 
                                      the standard installation paths. 

        Raises:
            Exception: If the rml file is not defined an exception is raised
        """
        if rml is not None:
            if isinstance(rml,RMLFile):
                self._rml = rml
            else: # assume that parameter is the file name as required for RMLFile
                self._rml = RMLFile(None,template=rml)
        else:
            raise Exception("rml file must be defined")
        self.path   = None
        self.prefix = 'RAYPy_Simulation'
        self._hide = hide
        self.analyze = True
        self._repeat = 1
        self.raypyng_analysis = True
        self.ray_path = ray_path


    @property
    def possible_exports(self):
        """A list of the files that can be exported by RAY-UI

        Returns:
            list: list of the names of the possible exports for RAY-UI
        """        
        self._possible_exports = ['AnglePhiDistribution',
                                'AnglePsiDistribution',
                                'BeamPropertiesPlotSnapshot',
                                'EnergyDistribution',
                                'FootprintAbsorbedRays',
                                'FootprintAllRays',
                                'FootprintOutgoingRays',
                                'FootprintPlotSnapshot',
                                'FootprintWastedRays',
                                'IntensityPlotSnapshot',
                                'IntensityX',
                                'IntensityYZ',
                                'PathlengthDistribution',
                                'RawRaysBeam',
                                'RawRaysIncoming',
                                'RawRaysOutgoing',
                                'ScalarBeamProperties',
                                'ScalarElementProperties'
                             ]
        return self._possible_exports
    
    @property
    def possible_exports_without_analysis(self):
        """A list of the files that can be exported by RAY-UI when the 
        analysis option is turned off

        Returns:
            list: list of the names of the possible exports for RAY-UI when analysis is off
        """        
        self._possible_exports_without_analysis = ['RawRaysIncoming',
                                'RawRaysOutgoing'
                             ]
        return self._possible_exports_without_analysis

    @property 
    def rml(self):
        """RMLFile object instantiated in init
        """        
        return self._rml

    @property 
    def simulation_name(self):
        """A string to append to the folder where the simulations will be executed.
        """        
        return self._simulation_name
    
    @simulation_name.setter
    def simulation_name(self,value):
        self._simulation_name = value

    @property 
    def analyze(self):
        """Turn on or off the RAY-UI analysis of the results. 
        The analysis of the results takes time, so turn it on only if needed

        Returns:
            bool: True: analysis on, False: analysis off
        """        
        return self._analyze
    
    @analyze.setter
    def analyze(self,value):
        if not isinstance(value, bool):
            raise ValueError ('Only bool are allowed')
        self._analyze = value

    @property 
    def raypyng_analysis(self):
        """Turn on or off the RAYPyNG analysis of the results. 

        Returns:
            bool: True: analysis on, False: analysis off
        """        
        return self._raypyng_analysis
    
    @raypyng_analysis.setter
    def raypyng_analysis(self,value):
        if not isinstance(value, bool):
            raise ValueError ('Only bool are allowed')
        self._raypyng_analysis = value
        
    @property 
    def repeat(self):
        """The simulations can be repeated an arbitrary number of times
        If the statitcs are not good enough using 2 millions of rays is suggested
        to repeat them instead of increasing the number of rays

        Returns:
            int: the number of repetition of the simulations, by default is 1
        """        
        return self._repeat
    
    @repeat.setter
    def repeat(self,value):
        if not isinstance(value, int):
            raise ValueError ('Only int are allowed')
        self._repeat = value

    @property 
    def path(self):
        """The path where to execute the simlations

        Returns:
            string: by default the path is the current path from which
            the program is executed
        """        
        return self._path

    @path.setter
    def path(self,value):
        if value == None:
            value=os.getcwd()
        if not isinstance(value, str):
            raise ValueError ('Only str are allowed')
        if not os.path.exists(value):
            raise ValueError('The path does not exist')
        self._path = value

    @property 
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self,value):
        if not isinstance(value, str):
            raise ValueError ('Only str are allowed')
        self._prefix = value

    @property 
    def exports(self):
        """The files to export once the simulation is complete.
        for a list of possible files check self.possible_exports
        and self.possible_exports_without_analysis.

        It is expeceted a list of dictionaries, and for each dictionary the key is the element 
        to be exported and the valuee are the files to be exported
        """        
        return self._exports

    @exports.setter
    def exports(self,value):
        if not isinstance(value, list):
            raise AssertionError ('The exports must be a list, while it is a '+str(type(value)), value)
        for d in value:
            if not isinstance(d,dict):
                raise AssertionError('The element of the list must be dictionaries, while I found a '+str(type(d)), d)
            for k in d.keys():
                if not isinstance(k,ObjectElement):
                    raise AssertionError('The keys of the dictionaries must be instance of ObjectElement, while ', k, 'is a ', str(type(k)))
                if isinstance(d[k], str):
                    if self.analyze:
                        possible_exports = self.possible_exports
                    else: 
                        possible_exports = self.possible_exports_without_analysis
                    if d[k] not in possible_exports:
                        raise AssertionError('It is not possible to export {}, check the spelling. The possible files to exports are {}'.format(d[k],possible_exports))
                elif isinstance(d[k], list):
                    for dd in d[k]:
                        if dd not in self.possible_exports and self._analyze==True:
                            raise AssertionError('It is not possible to export this file. The possible files to exports are ', self.possible_exports)
                        elif dd not in self.possible_exports_without_analysis and self._analyze==False:
                            raise AssertionError('It is not possible to export this file. The possible files to exports are ', self.possible_exports)

        self._exports = value
        self._exports_list = self.compose_exports_list(value, verbose=False)
        
    @property
    def params(self):
        """The parameters to scan, as a list of dictionaries.
        For each dictionary the keys are the parameters elements of the beamline, and the values are the 
        values to be assigned.
        """               
        return self.param

    @params.setter
    def params(self,value):
        if not isinstance(value, SimulationParams):
            #raise AssertionError('Params must be an instance of SimulationParams, while it is', type(params))
            sp = SimulationParams(self.rml)
            sp.params = value
            self.sp = sp
        else:
            self.sp = value
        _ = self.sp._extract_param(verbose=False)
        _ =self.sp._calc_loop()

    def save_parameters_to_file(self, dir):
        """save all the user input parameters to file. It takes the values
        from the SimulationParams class

        Args:
            dir (str): the folder where to save the parameters
        """        
        # do it first for indipendent parameters
        for i,p in enumerate(self.sp.ind_par):
            filename = str(p.get_full_path().lstrip("lab.beamline."))
            filename = "input_param_"+filename.replace(".", "_")
            filename += ".dat"
            np.savetxt(os.path.join(dir,filename),self.sp.ind_param_values[i])
        for i,p in enumerate(self.sp.dep_par):
            filename = str(p.get_full_path().lstrip("lab.beamline."))
            filename = "input_param_"+filename.replace(".", "_")
            filename += ".dat"
            np.savetxt(os.path.join(dir,filename),list(self.sp.dep_value_dependency[i].values()))
    
    def rml_list(self):
        """This function creates the folder structure and the rml files to simulate.
        It requires the param to be set. Useful if one wants to create the simulation files 
        for a manual check before starting the simulations.
        """            
        result = []
        self.sim_list_path = []
        self.sim_path = os.path.join(self.path, self.prefix+'_'+self.simulation_name)
        # check if simulation folder exists, otherwise create it
        if not os.path.exists(self.sim_path):
            os.makedirs(self.sim_path)
        self.save_parameters_to_file(self.sim_path)
        for r in range(0,self.repeat):
            sim_folder = os.path.join(self.sim_path,'round_'+str(r))
            if not os.path.exists(sim_folder):
                os.makedirs(sim_folder)
            for sim_n,param_set in enumerate(self.sp.params_list()):
                rml_path = os.path.join(sim_folder,str(sim_n)+'_'+self.simulation_name+'.rml')
                for param,value in param_set.items():
                    self.sp._write_value_to_param(param,value)
                self.rml.write(rml_path)
                self.sim_list_path.append(rml_path)
                # is this gonna create problems if I have millions of simulations?
                result.append(RMLFile(rml_path))

            # create csv file with simulations recap
            with open(os.path.join(sim_folder,'looper.csv'), 'w') as f:
                header = 'n '
                for par in self.sp.param_to_simulate:
                    #header = header + '\t'+str(par.id)#get_full_path().lstrip("lab.beamline."))
                    header = header + '\t'+str(par.get_full_path().lstrip("lab.beamline."))
                header += '\n'
                f.write(header)
                for ind,par in enumerate(self.sp.simulations_param_list):
                    line = ''
                    line += str(ind)+'\t'
                    for value in par:
                        line += str(value)+'\t'
                    f.write(line+'\n')  
        return result

    def compose_exports_list(self, exports_dict_list,/,verbose:bool=True):
        self.exports_list=[]
        for i, d in enumerate(self.exports):
                for obj in d.keys():
                    if isinstance(d[obj], str):
                        self.exports_list.append((obj.attributes().original()['name'], d[obj]))
                    elif isinstance(d[obj], list):
                        for l in d[obj]:
                            self.exports_list.append((obj.attributes().original()['name'], l))
                    else: 
                        raise ValueError('The exported param can be only str or list of str.')
        if verbose:
            print('The following will be exported:')
            for d in self.exports_list:
                print(d[0], d[1])

    def check_simulations(self,/,verbose:bool=True, force:bool=False):
        if force: return {k:v for k,v in enumerate(self.rml_list())}
        missing_simulations={}
        for ind,simulation in enumerate(self.rml_list()):
            folder = os.path.dirname(self.sim_list_path[ind])
            filename = os.path.basename(self.sim_list_path[ind])
            sim_number = filename.split("_")[0]
            for d in self.exports_list:
                export = sim_number+'_'+d[0]+'-'+d[1]+'.csv'
                csv = os.path.join(folder,export)
                if not os.path.exists(csv):
                    missing_simulations[ind]=simulation
                    break
        if verbose:
            print('I still have ', len(missing_simulations), 'simulations to do!')
            #print('missing_simulations',missing_simulations)
        return missing_simulations

    def run(self,recipe=None,/,multiprocessing=True, force=False):
        """This method starts the simulations. params and exports need to be defined.

        Args:
            recipe (SimulationRecipe, optional): If using a recipee pass it as a parameter. Defaults to None.
            multiprocessing (boolint, optional): If True all the cpus are used. If an integer n is provided, n cpus are used. Defaults to True.
            force (bool, optional): If True all the simlations are performed, even if the export files already exist. If False only the simlations for which are missing some exports are performed. Defaults to False.

        """             
        if recipe is not None:
            if isinstance(recipe,SimulationRecipe):
                self.params = recipe.params(self)
                self.exports = recipe.exports(self)
                self.simulation_name = recipe.simulation_name(self)
            else:
                raise TypeError("Unsupported type of the recipe!")

        filenames_hide_analyze = []
        exports = []
        missing_simulations= self.check_simulations(force=force).items()
        for ind,rml in missing_simulations:
            filename = os.path.basename(rml.filename)
            filenames_hide_analyze.append([rml.filename, self._hide, self._analyze, self.raypyng_analysis, self.ray_path])
            sim_index = int(filename[:filename.index("_")])
            exports.append(self.generate_export_params(sim_index,self.sim_list_path[ind]))
            rml.write()
        with RunPool(multiprocessing) as pool:
            pool.map(run_rml_func,zip(filenames_hide_analyze,exports))
        if len(missing_simulations) != 0 and self.analyze==False and self.raypyng_analysis==True:
            pp = PostProcess()
            pp.cleanup(self.sim_path, self.repeat, self.exports_list)
        return True

    def generate_export_params(self,simulation_index,rml):
        folder = os.path.dirname(rml)
        return [ (d[0], d[1], folder, str(simulation_index)+'_') for d in self.exports_list]
         
def run_rml_func(_tuple):
    filenames_hide_analyze,exports = _tuple
    rml_filename     = filenames_hide_analyze[0]
    hide             = filenames_hide_analyze[1]
    analyze          = filenames_hide_analyze[2]
    raypyng_analysis = filenames_hide_analyze[3]
    ray_path         = filenames_hide_analyze[4]
    runner = RayUIRunner(ray_path=ray_path,hide=hide)
    api    = RayUIAPI(runner)
    pp     = PostProcess()
    runner.run()
    api.load(rml_filename)
    api.trace(analyze=analyze)
    api.save(rml_filename)
    #print("DEBUG:: exports", exports)
    for e in exports:
        api.export(*e)
        if analyze==False and raypyng_analysis == True:
            pp.postprocess_RawRays(e[0], e[1], e[2], e[3], rml_filename)
    #time.sleep(0.1) # testing file creation issue
    try: 
        api.quit()
    except Exception as e:
        print("WARNING! Got exception while quitting ray, the error was:",e)
        pass
    #time.sleep(1) # testing file creation issue
    runner.kill()
    return None



###############################################################################
# XML helper class(es)
#
# Some code is based on the "untangle" package 
# see https://github.com/stchris/untangle and https://untangle.readthedocs.io/en/latest/
# fpr more details

###############################################################################


from .attributes import *
from .dictionaries import *
from .elements import *
from .parser import *


###############################################################################
def serialize(element:XmlElement,/,indent = "", filename=None):
    def serialize_children(strlist,element,indent, base_indent):
        if element.children() is not None:
            if len(element.children()) > 0:
                strlist.append('\n')
                for c in element.children():
                    strlist.append(serialize(c,indent=indent))
                strlist += [base_indent]
        if element.cdata is not None:
                strlist.append(element.cdata)
        return ''.join(strlist)

    strlist = []
    if element.is_root:
        serialize_children(strlist,element,"","")
    else:
        strlist = [indent+'<'+element.original_name()]
        if element.attributes() is not None and len(element.attributes())>0:
            strlist.append(' ')
            attrs = []
            for k,v in element.attributes().original().items():
                attrs+=[k+'="'+v+'"']
            strlist.append(" ".join(attrs))
        strlist.append('>')
    
        serialize_children(strlist,element,indent+"    ",base_indent=indent)

        strlist += ['</',element.original_name(),'>\n']
    result =  ''.join(strlist)
    return result



###############################################################################
# Generic superclasses for collections line dict() or list()

from collections.abc import MutableMapping,MutableSequence
import typing

###############################################################################
# based on the collections.UserList implementation from python 3.8.10
class MappedList(MutableSequence):
    """MappedList class implements a list which will apply the given 
    function to each value added to it and also will keep a track of original 
    values.
    """
    def __init__(self, func:typing.Callable=None, initlist=None):
        """Initializes instance of MappMappedListedDict class

        Args:
            func (typing.Callable, optional): function to be called on each
                new value added to the dictionary. Defaults to None.
            initlist (_type_, optional): Optional initilizer, any iterable object.
                Defaults to None.

        Raises:
            TypeError:  if suppled 'func' is not callable
        """
        if func is None:
            func = lambda x: x
        elif not callable(func):
            raise TypeError("callback parameter of 'MappedList' object must be callable")
        self.__func = func        

        self.__data = []
        self.__original_data = []

        if initlist is not None:
            self.extend(initlist)


    def __repr__(self): return repr(self.__data)
    def __lt__(self, other): return self.__data <  self.__cast(other)
    def __le__(self, other): return self.__data <= self.__cast(other)
    def __eq__(self, other): return self.__data == self.__cast(other)
    def __gt__(self, other): return self.__data >  self.__cast(other)
    def __ge__(self, other): return self.__data >= self.__cast(other)
    def __cast(self, other):
        return other.__data if isinstance(other, MappedList) else other
    def __contains__(self, item): return item in self.__data
    def __len__(self): return len(self.__data)
    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__class__(self.__data[i])
        else:
            return self.__data[i]
    def __setitem__(self, i, item):
        self.__original_data[i] = item
        self.__data[i] = self.__func(item)
    def __delitem__(self, i): 
        del self.__original_data[i]
        del self.__data[i]
    def __add__(self, other):
        if isinstance(other, MappedList):
            return self.__class__(self.__func,self.__original_data + other.__original_data)
        elif isinstance(other, type(self.__data)):
            return self.__class__(self.__func,self.__original_data + other)
        return self.__class__(self.__func,self.__original_data + list(other))
    def __radd__(self, other):
        if isinstance(other, MappedList):
            return self.__class__(self.__func,other.__original_data + self.__original_data)
        elif isinstance(other, type(self.__data)):
            return self.__class__(self.__func,other + self.__original_data)
        return self.__class__(self.__func,list(other) + self.__original_data)
    def __iadd__(self, other):
        if isinstance(other, MappedList):
            self.__data += other.__data
            self.__original_data += other.__original_data
        elif isinstance(other, type(self.__data)):
            self.__original_data += other
            self.__data += map(self.__func, other)
        else:
            self.__original_data += list(other)
            self.__data += map(self.__func, other)
        return self
    def __mul__(self, n):
        return self.__class__(self.__func, self.__orignal_data*n)
    __rmul__ = __mul__
    def __imul__(self, n):
        self.__original_data *= n
        self.__data *= n
        return self
    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        # Create a copy and avoid triggering descriptors
        inst.__dict__["_MappedList__data"] = self.__dict__["_MappedList__data"][:]
        inst.__dict__["_MappedList__original_data"] = self.__dict__["_MappedList__original_data"][:]
        inst.__dict__["_MappedList__func"] = self.__dict__["_MappedList__func"]
        return inst
    def append(self, item): 
        self.__original_data.append(item)
        self.__data.append(self.__func(item))
    def insert(self, i, item): 
        self.__original_data.insert(i, item)
        self.__data.insert(i, self.__func(item))
    def pop(self, i=-1): 
        self.__original_data.pop(i)
        return self.__data.pop(i)
    def remove(self, item): 
        self.__original_data.remove(item)
        self.__data.remove(item)
    def clear(self): 
        self.__original_data.clear()
        self.__data.clear()
    def copy(self): return self.__class__(self)
    def count(self, item): return self.__data.count(item)
    def index(self, item, *args): return self.__data.index(item, *args)
    def reverse(self): 
        self.__original_data.reverse()
        self.__data.reverse()
    def sort(self, /, *args, **kwds): 
        self.__original_data.sort(*args, **kwds)
        self.__data.sort(*args, **kwds)
    def extend(self, other):
        if isinstance(other, MappedList):
            self.__original_data.extend(other.__original_data)
            self.__data.extend(other.__data)
        else:
            self.__original_data.extend(other)
            self.__data.extend(map(self.__func,other))


###############################################################################
# based on the collections.UserDict implementation from python 3.8.10
class MappedDict(MutableMapping):
    """MappedDict class implements a dictionary which will apply the given 
    function to each value added to it and also will keep a track of original 
    values.
    """
    def __init__(self, func:typing.Callable=None,dict=None, **kwargs):
        """Initializes instance of MappedDict class

        Args:
            func (typing.Callable, optional): function to be called on each
                new value added to the dictionary. Defaults to None.
            dict (optional): Optional initilizer, any dict-like object. 
                Defaults to None.

        Raises:
            TypeError: if suppled 'func' is not callable
        """
        if func is None:
            func = lambda x: x
        elif not callable(func):
            raise TypeError("callback parameter of 'MappedDict' object must be callable")
        self.__func = func        
        self.__data = {}
        self.__original_data = {}

        if dict is not None:
            self.update(dict)
        if kwargs:
            self.update(kwargs)

    def original(self):
        """prodives acccess to the original, unmodified data

        Returns:
            dict: dictionary with original data
        """
        return self.__original_data

    def __len__(self): return len(self.__data)
    def __getitem__(self, key):
        if key in self.__data:
            return self.__data[key]
        if hasattr(self.__class__, "__missing__"):
            return self.__class__.__missing__(self, key)
        raise KeyError(key)
    def __setitem__(self, key, item):
        self.__original_data[key]=item
        self.__data[key] = self.__func(item)
    def __delitem__(self, key): 
        del self.__original_data[key]
        del self.__data[key]
    def __iter__(self):
        return iter(self.__data)

    # Modify __contains__ to work correctly when __missing__ is present
    def __contains__(self, key):
        return key in self.__data

    # Now, add the methods in dicts but not in MutableMapping
    def __repr__(self): return repr(self.__data)
    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        # Create a copy and avoid triggering descriptors
        inst.__dict__["_MappedDict__data"] = self.__dict__["_MappedDict__data"].copy()
        inst.__dict__["_MappedDict__original_data"] = self.__dict__["_MappedDict__original_data"].copy()
        inst.__dict__["_MappedDict__func"] = self.__dict__["_MappedDict__func"] 
        return inst

    def copy(self):
        if self.__class__ is MappedDict:
            return MappedDict(self.__func,self.__original_data.copy())
        import copy
        data = self.__data
        try:
            self.__data = {}
            c = copy.copy(self)
        finally:
            self.__data = data
        c.update(self)
        return c

    @classmethod
    def fromkeys(cls, iterable, value=None):
        d = cls()
        for key in iterable:
            d[key] = value
        return d




from .rml import ObjectElement
import numpy as np



class Simulate: pass

################################################################
class SimulationRecipe:
    def params(self,sim): return None
    def exports(self,sim): return None
    def simulation_name(self,sim): return None

################################################################
class ResolvingPower(SimulationRecipe):
    """Recipe for resolving power simulations.
    
    The eflectivity is automatically switched off for all elements, 
    the source if automatically recognized.

    Args:
            energy_range (np.array, list): the energies to simulate in eV
            exported_object (ObjectElement): the object to export
            source (ObjectElement, optional): the source object. 
                                              If None is provided, an automatic recogniton 
                                              of the source will be tried. Defaults to None.
            sim_folder (str, optional): the name of the simulation folder. If None, the rml 
                                        filename will be used. Defaults to None.
    """
    def __init__(self, energy_range:range, exported_object:ObjectElement,/, *args,source:ObjectElement=None,sim_folder:str=None):
        """
        Args:
            energy_range (np.array, list): the energies to simulate in eV
            exported_object (ObjectElement): the object to export
            source (ObjectElement, optional): the source object. 
                                              If None is provided, an automatic recogniton 
                                              of the source will be tried. Defaults to None.
            sim_folder (str, optional): the name of the simulation folder. If None, the rml 
                                        filename will be used. Defaults to None.
        """        
    
        if not isinstance(source, ObjectElement) and source != None:
            raise TypeError('The source must be an ObjectElement part of a beamline, while it is a', type(source))
        if not isinstance(energy_range, (range,np.ndarray)):
           raise TypeError('The energy_range must be an a ragne or a numpy array, while it is a', type(energy_range))
        if isinstance(exported_object, list):
            for exp_obj in exported_object:
                if not isinstance(exp_obj, ObjectElement):
                    raise TypeError('The exported_object must be an ObjectElement part of a beamline, while it is a', type(exported_object))
        else:
            if not isinstance(exported_object, ObjectElement):
                raise TypeError('The exported_object must be an ObjectElement part of a beamline, while it is a', type(exported_object))
            else:
                exported_object = [exported_object]

        self.source = source
        self.energy_range = energy_range
        self.exported_object = exported_object
        self.args = args
        self.sim_folder = sim_folder
    
    def params(self,sim:Simulate):
        params = []

        # find source and add to param with defined user energy range
        found_source = False
        if self.source == None:
            for oe in sim.rml.beamline.children():
                if hasattr(oe,"photonEnergy"):
                    self.source = oe
                    found_source = True
                    break        
            if found_source!=True:
                raise AttributeError('I did not find the source')        
        params.append({self.source.photonEnergy:self.energy_range})
        
        # handle free/optional arguments
        if self.args:
            for a in self.args:
                if not isinstance(a,dict):
                    raise TypeError('The args must be dictionaries, while I found a',type(a) )
                params.append(a)
        
        # turn reflectivity of all elements off, grating eff to 100%
        for oe in sim.rml.beamline.children():
                if hasattr(oe,"reflectivityType"):
                    params.append({oe.reflectivityType:0})

        # all done, return resulting params
        return params

    def exports(self,sim:Simulate):
        params = []
        if sim.analyze:
            export = "ScalarBeamProperties"
        else: 
            export = "RawRaysOutgoing"
        for exp_obj in self.exported_object:
            params.append({exp_obj:export})
        return params

    def simulation_name(self,sim:Simulate):
        if self.sim_folder is None:
            return 'RP'
        else: 
            return self.sim_folder
            
################################################################
class Flux(SimulationRecipe):
    """Recipe for flux simulations.
    
    The reflectivity is automatically switched on for all elements, 
    and the source is automatically identified.

    Args:
            energy_range (np.array, list): the energies to simulate in eV
            exported_object (ObjectElement): the object to export
            source (ObjectElement, optional): the source object. 
                                              If None is provided, an automatic recogniton 
                                              of the source will be tried. Defaults to None.
            sim_folder (str, optional): the name of the simulation folder. If None, the rml 
                                        filename will be used. Defaults to None.
    """
    def __init__(self, energy_range:range, exported_object:ObjectElement,/, *args,source:ObjectElement=None,sim_folder:str=None):
        """
        Args:
            energy_range (np.array, list): the energies to simulate in eV
            exported_object (ObjectElement): the object to export
            source (ObjectElement, optional): the source object. 
                                              If None is provided, an automatic recogniton 
                                              of the source will be tried. Defaults to None.
            sim_folder (str, optional): the name of the simulation folder. If None, the rml 
                                        filename will be used. Defaults to None.
        """        
    
        if not isinstance(source, ObjectElement) and source != None:
            raise TypeError('The source must be an ObjectElement part of a beamline, while it is a', type(source))
        if not isinstance(energy_range, (range,np.ndarray)):
           raise TypeError('The energy_range must be an a ragne or a numpy array, while it is a', type(energy_range))
        if isinstance(exported_object, list):
            for exp_obj in exported_object:
                if not isinstance(exp_obj, ObjectElement):
                    raise TypeError('The exported_object must be an ObjectElement part of a beamline, while it is a', type(exported_object))
        else:
            if not isinstance(exported_object, ObjectElement):
                raise TypeError('The exported_object must be an ObjectElement part of a beamline, while it is a', type(exported_object))
            else:
                exported_object = [exported_object]

        self.source = source
        self.energy_range = energy_range
        self.exported_object = exported_object
        self.args = args
        self.sim_folder = sim_folder
    
    def params(self,sim:Simulate):
        params = []

        # find source and add to param with defined user energy range
        found_source = False
        if self.source == None:
            for oe in sim.rml.beamline.children():
                if hasattr(oe,"photonEnergy"):
                    self.source = oe
                    found_source = True
                    break        
            if found_source!=True:
                raise AttributeError('I did not find the source')        
        params.append({self.source.photonEnergy:self.energy_range})
        
        # handle free/optional arguments
        if self.args:
            for a in self.args:
                if not isinstance(a,dict):
                    raise TypeError('The args must be dictionaries, while I found a',type(a) )
                params.append(a)
        
        # turn reflectivity of all elements on, grating eff to 100%
        for oe in sim.rml.beamline.children():
                if hasattr(oe,"reflectivityType"):
                    params.append({oe.reflectivityType:1})

        # all done, return resulting params
        return params

    def exports(self,sim:Simulate):
        params = []
        if sim.analyze:
            export = "ScalarBeamProperties"
        else: 
            export = "RawRaysOutgoing"
        for exp_obj in self.exported_object:
            params.append({exp_obj:export})
        return params

    def simulation_name(self,sim:Simulate):
        if self.sim_folder is None:
            return 'Flux'
        else: 
            return self.sim_folder

################################################################

class BeamWaist(SimulationRecipe):
    """Recipe for beamwaist simulations.
    
    At one defined energy export RawRaysOutgoing for all 
    optical elements, including image planes.

    Args:
            energy_range (np.array, list): the energies to simulate in eV
            source (ObjectElement, optional): the source object. If None is provided, 
                                              an automatic recogniton of the source 
                                              will be tried. Defaults to None.
            nrays (int): number of rays for the source
            sim_folder (str, optional): the name of the simulation folder. 
                                        If None, the rml filename will be used. 
                                        Defaults to None.
        
    """
    def __init__(self, energy:float,/,source:ObjectElement=None,nrays:int=None,sim_folder:str=None):
        """
        Args:
            energy_range (np.array, list): the energies to simulate in eV
            source (ObjectElement, optional): the source object. If None is provided, 
                                              an automatic recogniton of the source 
                                              will be tried. Defaults to None.
            nrays (int): number of rays for the source
            sim_folder (str, optional): the name of the simulation folder. 
                                        If None, the rml filename will be used. 
                                        Defaults to None.
        
        """        
    
        if not isinstance(energy, (int,float)):
           raise TypeError('The energy must be an a int or float, while it is a', type(energy))

        self.source = source
        self.energy = energy
        self.nrays  = nrays
        self.sim_folder = sim_folder
    
    def params(self,sim:Simulate):
        params = []

        # find source and add to param with defined user energy range
        found_source = False
        if self.source == None:
            for oe in sim.rml.beamline.children():
                if hasattr(oe,"photonEnergy"):
                    self.source = oe
                    found_source = True
                    break        
            if found_source!=True:
                raise AttributeError('I did not find the source')        
        params.append({self.source.photonEnergy:self.energy})
        
        for oe in sim.rml.beamline.children():
                for par in oe:
                    try:
                        params.append({par.reflectivityType:0})
                    except:
                        pass

        # all done, return resulting params
        return params

    def exports(self,sim:Simulate):
        oe_list=[]
        for oe in sim.rml.beamline.children():
            oe_list.append(oe)
        exports = []
        for oe in oe_list:
            exports.append({oe:'RawRaysOutgoing'})
            # print('DEBUG:: exported oe', oe.name)
        return exports

    def simulation_name(self,sim:Simulate):
        if self.sim_folder is None:
            return 'Beamwaist'
        else: 
            return self.sim_folder

################################################################

    



###############################################################################
class XmlAttribute:
    def __init__(self,xmlvalue=None) -> None:
        if xmlvalue is not None:
            self.__value=xmlvalue

    def set(self,value):
        self.__value = value

    def get(self):
        return self.__value

    def __str__(self):
        return str(self.__value)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self}')"

###############################################################################
class XmlMappedAttribute(XmlAttribute):
    def __init__(self,xmlvalue=None,map=None) -> None:
        if map is None:
            raise ValueError(f"map parameter is required and can not be 'None'")
        self.__map = map
        super().__init__(xmlvalue)

        if xmlvalue in self.__map:
            self.set(self.__map[xmlvalue] )
        else:
            raise ValueError(f"Invalid value for the XmlAttribute: {xmlvalue}")


    def __str__(self):
        for k,v in self.__map.items():
            if v==self.get():
                return k
        raise ValueError(f"Can not map to bool: {self.get()}")

    def __repr__(self):
        return f"{self.__class__.__name__}('{self}',map={self.__map})"


###############################################################################
class XmlBoolAttribute(XmlMappedAttribute):
    def __init__(self,xmlvalue=None,true='True',false='False') -> None:
        m = {true:True,false:False}
        super().__init__(xmlvalue,m)




from multiprocessing import parent_process
from xml.sax import make_parser, handler
from .elements import *
from .dictionaries import *


###############################################################################
global_known_classes = {} # part of the development code
class Handler(handler.ContentHandler):
    
    #####################################
    def __init__(self,/,known_classes=None):
        self.root = XmlElement(None, None)
        self.root.is_root = True
        self.elements = []
        if known_classes is None:
            self._known_classes = global_known_classes
        else:
            self._known_classes = known_classes


    #####################################
    def startElement(self, name, attributes):
        """called on the start of an element in non-namespace mode.

        Args:
            name (_type_): _description_
            attributes (_type_): _description_
        """
        #print("DEBUG::startElement::name=",name)

        # store attributes in a dictionary
        # local copy is nessesary, input object can change and we should not 
        # save reference to it

        # Shall it be moved into the 
        attrs = SafeValueDict()
        for k, v in attributes.items():
            attrs[k] = v#self.protectName(v)
        
        # create a new element
        if len(self.elements) > 0:
            parent = self.elements[-1]
        else:
            parent = self.root

        if name in self._known_classes.keys():
            element = self._known_classes[name](name, attrs, parent=parent)
        else:
            element = XmlElement(name, attrs,paren=parent)
        # and add it to the known element list
        parent.add_child(element)
        self.elements.append(element)

    #####################################
    def endElement(self, name):
        """called the end of an element in non-namespace mode.

        Args:
            name (_type_): _description_
        """
        self.elements.pop()

    #####################################
    def characters(self, cdata):
        """adds character data to the current element

        Args:
            cdata (_type_): _description_
        """
        self.elements[-1].add_cdata(cdata.strip())

###############################################################################
def parse(filename:str, /, known_classes = None, **parser_features)->XmlElement:
    if filename is None:
        raise ValueError("parse() takes a filename")
    parser = make_parser()
    for feature, value in parser_features.items():
        parser.setFeature(getattr(handler, feature), value)
    sax_handler = Handler(known_classes=known_classes)
    parser.setContentHandler(sax_handler)
    parser.parse(filename)
    return sax_handler.root




from ..collections import MappedList, MappedDict
import keyword


###############################################################################
def sanitizeName(name:str)->str:
    """convert name into python attribute safe name

    Args:
        name (str): _description_

    Returns:
        str: _description_
    """

    if name is None:
        return None

    # repalce special characters with _
    name = name.replace("-", "_")
    name = name.replace(".", "_")
    name = name.replace(":", "_")

    # delete spaces
    name = name.replace(" ", "")

    # adding trailing _ for keywords
    if keyword.iskeyword(name):
        name += "_"
    return name

###############################################################################
class SafeValueDict(MappedDict):
    def __init__(self, dict=None, **kwargs):
        super().__init__(sanitizeName, dict, **kwargs)

###############################################################################
class SafeValueList(MappedList):
    def __init__(self, initlist=None):
        super().__init__(sanitizeName, initlist)


